/***************************************************************************
 *   Copyright (C) 2008 by msm net ingenieurbuero meissner   *
 *   service@msm-net.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/*
TODO:
	- time not before
	- time not after
	- status
*/

#include "main.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <dirent.h>


enum CERTOBJECTS_type {
	KEY = 1,
	REQ,
	ROOTCA,
	SUBCA,
	CERT,
	CRL
};

char *CERTOBJECTS_endings[] = { NULL, ".key", ".req", ".rca", ".sca", ".cert", ".crl", NULL };

enum CERTOBJECTS_yesno {
	no = 0,
	yes
};

enum CERTOBJECTS_status {
	not_yet_valid = 1,
	valid,
	expired,
	expire_in_1,
	expire_in_10,
	expire_in_30,
	revoked,
	invalid
};

struct CERTOBJECTS_t {
	struct CERTOBJECTS_t *next;
	int count;
	enum CERTOBJECTS_type type;
	char *data;
	enum CERTOBJECTS_yesno	smartcard;
	char *filename;
	char *pubkey_hash; 			// ende Key
	enum CERTOBJECTS_yesno	has_key;
	char *subject_hash;
	char *subject;
	char *cn;				// ende Req
	char *issuer_hash;
	char *issuer;
	time_t time_not_after;
	enum CERTOBJECTS_status	status;		// ende Crl
	time_t time_not_before;
	char *serial;
	char *keyid_subject;
	char *keyid_issuer;
} CERTOBJECTS_dummy;

struct CERTOBJECTS_t *CERTOBJECTS = NULL;

// common name von String extrahieren
char *CERT_extract_cn(char *subject) {
     char *p_start, *p_end;

     if(	NULL != (p_start = strstr(subject, "/CN="))
     	&&	NULL != (p_end = strstr(p_start + 1, "/")))
		return strndup(p_start + 4, p_end - p_start - 4);

	 return strdup(".");
}

// info von Req/Cert/Crl holen (openssl)
char *CERT_get_info(enum CERTOBJECTS_type type, char *data, char *info) {
	char *text_out, *text_err;
	char *argv[] = { BIN_OPENSSL, NULL, "-noout", info, NULL };

	switch(type) {
		case REQ:	argv[1] = "req";
			break;
		case ROOTCA:
		case SUBCA:
		case CERT:	argv[1] = "x509";
			break;
		case CRL:	argv[1] = "crl";
			break;
		default:	return strdup(".");
			break;
	}

//printf("<br>%s %s %d (%s):<br>\n%s\n", __FUNCTION__, __FILE__, __LINE__, info, data);

	if(exec_command(data, &text_out, &text_err, argv, NULL) || text_out == NULL) {
		if(NULL != text_err)
			set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv[0], text_err);
		return strdup(".");
	}

	if(NULL != text_err)
		free(text_err);

	return text_out;
}

// subject von Req/Cert holen (openssl)
char *CERT_get_subject(enum CERTOBJECTS_type type, char *data) {
	char *info, *text, *output;

	output = CERT_get_info(type, data, "-subject");

	if(		NULL != output
		&&	NULL != (text = strstr(output, "subject="))
		&&	NULL != (text = strstr(text, "/"))) {

		if(NULL != index(text, '\n'))
			*index(text, '\n') = 0;

		info = strdup(text);
		free(output);
		return info;
	}

	return strdup(".");
}

// issuer von Req/Cert holen (openssl)
char *CERT_get_issuer(enum CERTOBJECTS_type type, char *data) {
	char *info, *text, *output;

	output = CERT_get_info(type, data, "-issuer");

	if(		NULL != output
		&&	NULL != (text = strstr(output, "issuer="))
		&&	NULL != (text = strstr(text, "/"))) {

		if(NULL != index(text, '\n'))
			*index(text, '\n') = 0;

		info = strdup(text);
		free(output);
		return info;
	}

	return strdup(".");
}

// Zeit bis wann gueltig Cert/Crl
time_t CERT_time_not_after(enum CERTOBJECTS_type type, char *data) {
	time_t x = time(NULL);

	return x;
}

// Zeit ab wann gueltig Cert
time_t CERT_time_not_before(enum CERTOBJECTS_type type, char *data) {
	time_t x = time(NULL);

	return x;
}

// Keyid des subjects
char *CERT_get_keyid_subject(enum CERTOBJECTS_type type, char *data) {
	char *info, *text, *output;

	output = CERT_get_info(type, data, "-text");

	if(		NULL != output
		&&	NULL != (text = strstr(output, "X509v3 Subject Key Identifier:"))
		&&	NULL != (text = index(text, '\n'))) {

		while(*(++text) == ' ' || *text == '\t');

		if(NULL != index(text, '\n'))
			*index(text, '\n') = 0;

		info = strdup(text);
		free(output);
//printf("X509v3 Subject Key Identifier: %s<br>\n", info);		
		return info;
	}

	return strdup(".");
}

// Keyid des issuers
char *CERT_get_keyid_issuer(enum CERTOBJECTS_type type, char *data) {
	char *info, *text, *output;

	output = CERT_get_info(type, data, "-text");

	if(		NULL != output
		&&	NULL != (text = strstr(output, "X509v3 Authority Key Identifier:"))
		&&	NULL != (text = strstr(text, "keyid:"))) {

		text += 6;

		if(NULL != index(text, '\n'))
			*index(text, '\n') = 0;

		info = strdup(text);
		free(output);
//printf("X509v3 Authority Key Identifier: %s<br>\n", info);				
		return info;
	}

	return strdup(".");
}

// typ nach CN sortieren
void CERT_sort_type_by_cn(enum CERTOBJECTS_type type) {
	struct CERTOBJECTS_t *certobject_start, *certobject_akt, *certobject_mem;
	struct CERTOBJECTS_t **certobject_p1, **certobject_p2, **certobject_p3;
	int changed;

	if(NULL != (certobject_start = CERTOBJECTS)) {

		while(NULL != certobject_start->next) {

			if(certobject_start->next->type == type)
				break;

			certobject_start = certobject_start->next;
		}

		do {
			changed = 0;
			certobject_akt = certobject_start;

			while(		certobject_akt->next != NULL
					&&	certobject_akt->next->type == type
					&&	certobject_akt->next->next != NULL
					&&	certobject_akt->next->next->type == type ) {

				if(	NULL != certobject_akt->next->cn
					&&	NULL != certobject_akt->next->next->cn ) {

					if(0 < strcmp(certobject_akt->next->cn, certobject_akt->next->next->cn)) {
						changed++;
						certobject_p1 = &certobject_akt->next;
						certobject_p2 = &certobject_akt->next->next;
						certobject_p3 = &certobject_akt->next->next->next;
						certobject_mem = *certobject_p1;
						*certobject_p1 = *certobject_p2;
						*certobject_p2 = *certobject_p3;
						*certobject_p3 = certobject_mem;
					}

				}

				certobject_akt = certobject_akt->next;

			}

		} while (changed);

	}
	return;
}

// Liste laden
int CERT_read_objects(enum CERTOBJECTS_type type, char* path) {
	DIR *directory;
	struct dirent *directory_structur;
	struct CERTOBJECTS_t *new_certobject = NULL, *last_certobject = NULL, *search_certobject;
	char *p_start, *p_end, *p_search;
	char full_path_name[1024];

	if(NULL == CERTOBJECTS) {
		CERTOBJECTS = calloc(1, sizeof(struct CERTOBJECTS_t));
		CERTOBJECTS->next = NULL;
		CERTOBJECTS->data = NULL;
		CERTOBJECTS->filename = NULL;
		CERTOBJECTS->pubkey_hash = strdup("");
		CERTOBJECTS->subject_hash = NULL;
		CERTOBJECTS->subject = NULL;
		CERTOBJECTS->cn = NULL;
		CERTOBJECTS->issuer_hash = NULL;
		CERTOBJECTS->issuer = NULL;
		CERTOBJECTS->serial = NULL;
		CERTOBJECTS->keyid_subject = NULL;
		CERTOBJECTS->keyid_issuer = NULL;
	}

	last_certobject = CERTOBJECTS;

	while(NULL != last_certobject->next)
		last_certobject = last_certobject->next;
	
	// Oeffne Zertifikats-Verzeichnis
	if(NULL != (directory = opendir(path))) {
		// lese alle Dateinamen

		while(NULL != (directory_structur = readdir(directory))) {
			// voller Pfadname
			snprintf(full_path_name, sizeof(full_path_name) - 1, "%s/%s", path, directory_structur->d_name);

			if(NULL != strstr(directory_structur->d_name, ".pem"))
				unlink(full_path_name);
   
			// wenn Datei vom aktuellen Typ ist
			if(NULL != strstr(directory_structur->d_name, CERTOBJECTS_endings[type])) {
				// erzeuge neuen Eintrag
// printf("%s/%s<br>\n",path, directory_structur->d_name);

				if(NULL == (new_certobject = calloc(1, sizeof(struct CERTOBJECTS_t))))
					return EXIT_FAILURE;
//printf("<br><br>%s %s %d:<br>\n%s\n", __FUNCTION__, __FILE__, __LINE__, directory_structur->d_name);

				last_certobject->next = new_certobject;
				last_certobject = new_certobject;
				// Zeiger auf NULL
				new_certobject->next = NULL;
				new_certobject->pubkey_hash = NULL;
				new_certobject->subject_hash = NULL;
				new_certobject->subject = NULL;
				new_certobject->cn = NULL;
				new_certobject->issuer_hash = NULL;
				new_certobject->issuer = NULL;
				new_certobject->serial = NULL;
				new_certobject->keyid_subject = NULL;
				new_certobject->keyid_issuer = NULL;

				// lese Daten
				if(read_file_in_var(full_path_name, "_data_"))
					 return EXIT_FAILURE;

				new_certobject->data = strdup(getvar("_data_"));
				
				// schreibe allgemeine Daten
				new_certobject->type = type;
				new_certobject->smartcard = no;
				new_certobject->filename = strdup(directory_structur->d_name);

				if(NULL != index(new_certobject->filename, '_'))
					new_certobject->smartcard = yes;

				// lade Hashwert des oeffentlichen Schluessels
				if(NULL == (p_end = strstr(new_certobject->filename, "-")))
					if(NULL == (p_end = strstr(new_certobject->filename, ".")))
						return EXIT_FAILURE;

				new_certobject->pubkey_hash = strndup(new_certobject->filename, p_end - new_certobject->filename);
				p_start = ++p_end;
					// wenn Schluessel dann hier ende

				if(type == KEY)
					continue;

				// lade Hashwert des Subjects
				if(NULL == (p_end = strstr(p_start, "-")))
					if(NULL == (p_end = strstr(p_start, ".")))
						return EXIT_FAILURE;

				new_certobject->subject_hash = strndup(p_start, p_end - p_start);
				p_start = ++p_end;

				// lade Subject
				if(type != CRL) {
					new_certobject->subject = CERT_get_subject(new_certobject->type, new_certobject->data);
					// ermittle Allgemeinen Namen
					new_certobject->cn = CERT_extract_cn(new_certobject->subject);
				}

				// pruefe ob Schluessel vorhanden
				new_certobject->has_key = no;
				search_certobject = CERTOBJECTS;
				do {
					if(!strcmp(search_certobject->pubkey_hash, new_certobject->pubkey_hash)) {
						new_certobject->has_key = yes;
						break;
					}
				}
				while (		NULL != (search_certobject = search_certobject->next)
						&&	search_certobject->type == KEY);

				// wenn Anforderung dann hier ende
				if(type == REQ)
					continue;

				// lade Hashwert des Issuers
				if(NULL == (p_end = strstr(p_start, "-")))
					if(NULL == (p_end = strstr(p_start, ".")))
						return EXIT_FAILURE;

				if(NULL != (p_search= strstr(p_start, "-")) && p_search < p_end)
					p_end = p_search;

				new_certobject->issuer_hash = strndup(p_start, p_end - p_start);
				p_start = ++p_end;
				// lade Issuer
				new_certobject->issuer = CERT_get_issuer(new_certobject->type, new_certobject->data);

				// Gueltigkeitsdauer holen
				new_certobject->time_not_after = CERT_time_not_after(new_certobject->type, new_certobject->data);

				// Status 1. Teil (bis wann gueltig / verifizieren)
				new_certobject->status = valid;

				// issuer-KeyId
				new_certobject->keyid_issuer = CERT_get_keyid_issuer(new_certobject->type, new_certobject->data);

				// wenn Sperrliste dann hier ende
				if(type == CRL)
					continue;

				// ermitteln ab wann gueltig
				new_certobject->time_not_before = CERT_time_not_before(new_certobject->type, new_certobject->data);

				// Status 2. Teil (ab wann gueltig / ist CA fuer Zertifikat vorhanden)
				new_certobject->status = valid;

				// Seriennummer
				if(NULL == (p_end = strstr(p_start, "-")))
					if(NULL == (p_end = strstr(p_start, ".")))
						return EXIT_FAILURE;

				new_certobject->serial = strndup(p_start, p_end - p_start);

				// subject-KeyId
				new_certobject->keyid_subject = CERT_get_keyid_subject(new_certobject->type, new_certobject->data);
			}
		}
		// schliesse Verzeichnis-Handle
		closedir(directory);
	}

	return EXIT_SUCCESS;
}


// Zertifikat-Object-Liste erzeugen
int CERT_create_objectlist(void) {
	enum CERTOBJECTS_type type;
/*	char *argv_readcard[] = { BIN_READSMARTCARD, CERT_PATH_SC, NULL };
	char *text_out, *text_err;

    if(exec_command(NULL, &text_out, &text_err, argv_readcard, NULL)) {
		if(NULL != text_err)
			set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv_readcard[0], text_err);
		return EXIT_FAILURE;
	}
*/
	for(type = KEY; type <= CRL; type++) {
		CERT_read_objects(type, CERT_PATH);
		CERT_read_objects(type, CERT_PATH_SC);
	}

	CERT_sort_type_by_cn(ROOTCA);
	CERT_sort_type_by_cn(SUBCA);

	return EXIT_SUCCESS;
}


int CERT_clear_objectlist(void) {
	struct CERTOBJECTS_t *last;

	while(NULL != (last = CERTOBJECTS)) {
		 if(NULL != CERTOBJECTS->data)
		 	free(CERTOBJECTS->data);
//printf("<br><br>filename<br>%s\n", CERTOBJECTS->filename);

		 if(NULL != CERTOBJECTS->filename)
		 	free(CERTOBJECTS->filename);
//printf("<br><br>pubkey_hash<br>%s\n", CERTOBJECTS->pubkey_hash);

		 if(NULL != CERTOBJECTS->pubkey_hash)
		 	free(CERTOBJECTS->pubkey_hash);
//printf("<br><br>subject_hash<br>%s\n", CERTOBJECTS->subject_hash);

		 if(NULL != CERTOBJECTS->subject_hash)
		 	free(CERTOBJECTS->subject_hash);
//printf("<br><br>subject<br>%s\n", CERTOBJECTS->subject);

		 if(NULL != CERTOBJECTS->subject)
		 	free(CERTOBJECTS->subject);
//printf("<br><br>cn<br>%s\n", CERTOBJECTS->cn);

		 if(NULL != CERTOBJECTS->cn)
		 	free(CERTOBJECTS->cn);
//printf("<br><br>issuer_hash<br>%s\n", CERTOBJECTS->issuer_hash);

		 if(NULL != CERTOBJECTS->issuer_hash)
		 	free(CERTOBJECTS->issuer_hash);
//printf("<br><br>issuer<br>%s\n", CERTOBJECTS->issuer);

		 if(NULL != CERTOBJECTS->issuer)
		 	free(CERTOBJECTS->issuer);
//printf("<br><br>serial<br>%s\n", CERTOBJECTS->serial);

		 if(NULL != CERTOBJECTS->serial)
		 	free(CERTOBJECTS->serial);
//printf("<br><br>keyid_subject<br>%s\n", CERTOBJECTS->keyid_subject);

		 if(NULL != CERTOBJECTS->keyid_subject)
		 	free(CERTOBJECTS->keyid_subject);
//printf("<br><br>keyid_issuer<br>%s\n", CERTOBJECTS->keyid_issuer);

		 if(NULL != CERTOBJECTS->keyid_issuer)
		 	free(CERTOBJECTS->keyid_issuer);

		 CERTOBJECTS = CERTOBJECTS->next;
		 free(last);
	}

	return EXIT_SUCCESS;
}



int CERT_create_optionlist_certs(char **p_buffer_main, char *selected, int without_keys_also) {
	enum CERTOBJECTS_type type_main;
	struct CERTOBJECTS_t *search_certobject_main, *search_certobject_sub;
	char *p_buffer_sub = NULL;
	char buffer[4096], id[5] = {0};

	if(NULL != selected && selected == strstr(selected, "%smartcard0:"))
		snprintf(id, sizeof(id), "_%d", atoi(selected += 12));

	if(NULL == CERTOBJECTS)
		CERT_create_objectlist();
	 
	 // Liste unterzeichungs-faehiger CAs zusammenstellen
	if(NULL != CERTOBJECTS) {
		*p_buffer_main = NULL;
		for(type_main = ROOTCA; type_main <= CERT; type_main++) {
			search_certobject_main = CERTOBJECTS;

			do {

				// RootCAs suchen
				if(		search_certobject_main->type == type_main
					&&	(search_certobject_main->type != CERT || !search_certobject_main->count)) {

					snprintf(buffer, sizeof(buffer), "<option disabled value=\"%s\" title=\"%s\">%s [%s", \
						search_certobject_main->filename, search_certobject_main->subject, \
						search_certobject_main->cn, search_certobject_main->serial);
//printf("MAIN=\"%s\" title=\"%s\"%s<br>\n",search_certobject_main->filename, search_certobject_main->subject, search_certobject_main->cn);

					if(search_certobject_main->smartcard)
							strcat(buffer, " s");

					strcat(buffer, "]");
					search_certobject_main->count++;

					if(*p_buffer_main != NULL) {
						*p_buffer_main = realloc(*p_buffer_main, strlen(*p_buffer_main) + strlen(buffer) + 1);
						strcat(*p_buffer_main, buffer);
					}

					else {
						*p_buffer_main = malloc(strlen(buffer) + 1);
						strcpy(*p_buffer_main, buffer);
					}
					// Identzertifikate fuer CA suchen
					search_certobject_sub = CERTOBJECTS;

					while(NULL != search_certobject_sub->next) {
						search_certobject_sub = search_certobject_sub->next;
						if(		search_certobject_sub->type == CERT
							&&	!strcmp(search_certobject_sub->issuer_hash, search_certobject_main->subject_hash)
							&&	!strcmp(search_certobject_sub->keyid_issuer, search_certobject_main->keyid_subject)
							&&	(search_certobject_sub->has_key || without_keys_also)
						) {

							if(		NULL != selected
								&&	(		(!*id && !strcmp(selected, search_certobject_sub->filename))
										||	(*id && NULL != strstr(search_certobject_sub->filename, id))
									)
							)
								snprintf(buffer, sizeof(buffer), "<option selected value=\"%s\" title=\"%s\">&nbsp;&nbsp;&nbsp;%s [%s",
									search_certobject_sub->filename, search_certobject_sub->subject, \
									search_certobject_sub->cn, search_certobject_sub->serial);
							else
								snprintf(buffer, sizeof(buffer), "<option value=\"%s\" title=\"%s\">&nbsp;&nbsp;&nbsp;%s [%s",
									search_certobject_sub->filename, search_certobject_sub->subject, \
									search_certobject_sub->cn, search_certobject_sub->serial);
//printf("SUB=\"%s\" title=\"%s\"%s<br>\n",search_certobject_sub->filename, search_certobject_sub->subject, search_certobject_sub->cn);

							if(search_certobject_sub->smartcard)
								strcat(buffer, " s");

							strcat(buffer, "]");
							search_certobject_sub->count++;

							if(p_buffer_sub != NULL) {
								p_buffer_sub = realloc(p_buffer_sub, strlen(p_buffer_sub) + strlen(buffer) + 1);
								strcat(p_buffer_sub, buffer);
							}

							else {
								p_buffer_sub = malloc(strlen(buffer) + 1);
								strcpy(p_buffer_sub, buffer);
							}
						}
					}

					if(NULL != p_buffer_sub) {
						sort_optionlist(p_buffer_sub);
						*p_buffer_main = realloc(*p_buffer_main, strlen(*p_buffer_main) + strlen(p_buffer_sub) + 1);
						strcat(*p_buffer_main, p_buffer_sub);
						free(p_buffer_sub);
						p_buffer_sub = NULL;
					}
				}

			} while(NULL != (search_certobject_main = search_certobject_main->next));

		}
	}

	return EXIT_SUCCESS;
}


int CERT_create_optionlist_cas(char **p_buffer_main, char *selected, int by_subject) {
	enum CERTOBJECTS_type type_main;
	struct CERTOBJECTS_t *search_certobject_main;
	char buffer[4096], id[5] = {0};

	if(NULL == CERTOBJECTS)
	CERT_create_objectlist();

	if(NULL != selected && selected == strstr(selected, "%smartcard0:"))
		snprintf(id, sizeof(id), "_%d", atoi(selected += 12));

	if(NULL != CERTOBJECTS) {
		*p_buffer_main = NULL;

		for(type_main = ROOTCA; type_main <= CERT; type_main++) {
			search_certobject_main = CERTOBJECTS;

			do {
				// RootCAs suchen
				if(		search_certobject_main->type == type_main
					&&	(search_certobject_main->type != CERT || !search_certobject_main->count)) {

					if(	NULL != selected
						&&	(	(	!by_subject
								&&	(
									(!*id && !strcmp(selected, search_certobject_main->filename))
								||	(*id && NULL != strstr(search_certobject_main->filename, id))
									)
							)
						||	(by_subject && !strcmp(selected, search_certobject_main->subject))
						)
					)
						snprintf(buffer, sizeof(buffer), "<option selected value=\"%s\" title=\"%s\">%s [%s", \
							search_certobject_main->filename, search_certobject_main->subject, \
							search_certobject_main->cn, search_certobject_main->serial);

					else
						snprintf(buffer, sizeof(buffer), "<option value=\"%s\" title=\"%s\">%s [%s", \
							search_certobject_main->filename, search_certobject_main->subject, \
							search_certobject_main->cn, search_certobject_main->serial);
//printf("MAIN=\"%s\" title=\"%s\"%s<br>\n",search_certobject_main->filename, search_certobject_main->subject, search_certobject_main->cn);

					if(search_certobject_main->smartcard)
							strcat(buffer, " s");

					strcat(buffer, "]");
					search_certobject_main->count++;

					if(*p_buffer_main != NULL) {
						*p_buffer_main = realloc(*p_buffer_main, strlen(*p_buffer_main) + strlen(buffer) + 1);
						strcat(*p_buffer_main, buffer);
					}

					else {
						*p_buffer_main = malloc(strlen(buffer) + 1);
						strcpy(*p_buffer_main, buffer);
					}

				}

			} while(NULL != (search_certobject_main = search_certobject_main->next));

		}
	}

	return EXIT_SUCCESS;
}


