/***************************************************************************
 *   Copyright (C) 2008 by msm net ingenieurbuero meissner   *
 *   service@msm-net.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <dirent.h>

#include "main.h"
#include "CERT.h"


extern int certcertlist(void);

int certcertimport(void) {
	char text_in[8192], filename[1024], keypass[100];
	char *text_out = NULL, *text_err = NULL;
	char *argv_ssl[] = { BIN_OPENSSL, NULL };
	char *argv_cert_key[] = { BIN_OPENSSL, "x509", "-noout", "-pubkey", NULL };
	char *argv_key_key[] = { BIN_OPENSSL, "rsa", "-pubout", "-passin", keypass, NULL };
	char *argv_insert[] = { BIN_INSERTCERTS, PATH_TEMP "/tmpcert", NULL, NULL };
	char *p_start, *p_end;
	int count = 0;
	FILE *p_file;

	if(NULL != getvar("execute")) {
		if(NULL == getvar("p12file_size") || 0 == atoi(getvar("p12file_size")) || NULL == getvar("p12file_pointer")) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 31);
			return EXIT_FAILURE;
		}

		if(NULL == getvar("password")) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 32);
			return EXIT_FAILURE;
		}

		if(NULL == getvar("keypassword") || NULL == getvar("keypassword1")) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 33);
			return EXIT_FAILURE;
		}
		
		// file local speichern
		if(NULL != (p_file = fopen(PATH_TEMP "/tmp12.p12", "w"))) {
			 fwrite(getvar("p12file_pointer"), atoi(getvar("p12file_size")), 1, p_file);
			 fclose(p_file);
		}
		else {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 34);
			return EXIT_FAILURE;
		}

		// ident-cert extrahieren
		snprintf(text_in, sizeof(text_in) - 1, "pkcs12 -clcerts -nokeys -in " PATH_TEMP "/tmp12.p12 -passin \"pass:%s\"\n", getvar("password"));

		if(exec_command(text_in, &text_out, &text_err, argv_ssl, NULL)) {
			set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv_ssl[0], text_err);
			return EXIT_FAILURE;
		}

		setvar(text_out, "_ident_");
		p_start = getvar("_ident_");
		count = 0;
		
		while(p_start != NULL && NULL != (p_start = strstr(p_start, "-----BEGIN"))) {
			count++;
			p_start++;
		}

		if(count != 1) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 35);
			return EXIT_FAILURE;
		}

		if(exec_command(getvar("_ident_"), &text_out, &text_err, argv_cert_key, NULL)) {
			set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv_cert_key[0], text_err);
			return EXIT_FAILURE;
		}

		setvar(text_out, "_ident_pk_");
		
        // privat key extrahieren
		snprintf(text_in, sizeof(text_in) - 1, "pkcs12 -nocerts -in " PATH_TEMP "/tmp12.p12 -passin \"pass:%s\" -passout \"pass:%s\"\n", \
			getvar("password"), getvar("keypassword"));

		if(exec_command(text_in, &text_out, &text_err, argv_ssl, NULL)) {
			set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv_ssl[0], text_err);
			return EXIT_FAILURE;
		}

		setvar(text_out, "_key_");
		p_start = getvar("_key_");
		count = 0;

		while(p_start != NULL && NULL != (p_start = strstr(p_start, "-----BEGIN"))) {
			count++;
			p_start++;
		}

		if(count != 1) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 36);
			return EXIT_FAILURE;
		}

		snprintf(keypass, sizeof(keypass) - 1, "pass:%s", getvar("keypassword"));

		if(exec_command(getvar("_key_"), &text_out, &text_err, argv_key_key, NULL)) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv_key_key[0], text_err);
			return EXIT_FAILURE;
		}

		setvar(text_out, "_key_pk_");

		// public keys vergleichen
		if(NULL == getvar("_ident_pk_") || NULL == getvar("_key_pk_") || strcmp(getvar("_ident_pk_"), getvar("_key_pk_"))) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 37);
			return EXIT_FAILURE;
		}

        // speichern
		save_var_as_file(PATH_TEMP "/tmpcert.cert", "_ident_");
		save_var_as_file(PATH_TEMP "/tmpcert.key", "_key_");
		
		// CA-Zertifikate laden
		snprintf(text_in, sizeof(text_in) - 1, "pkcs12 -cacerts -nokeys -in " PATH_TEMP "/tmp12.p12 -passin \"pass:%s\"\n", getvar("password"));

		if(exec_command(text_in, &text_out, &text_err, argv_ssl, NULL)) {
			set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv_ssl[0], text_err);
			return EXIT_FAILURE;
		}

		setvar(text_out, "_ca_");
		p_start = getvar("_ca_");
		count = 0;

		while(NULL != p_start && NULL != (p_start = strstr(p_start, "-----BEGIN"))) {
			if(NULL != (p_end = strstr(p_start, "-----\n")) && NULL != (p_end = strstr(p_end + 1, "-----\n"))) {
				*(p_end += 6) = 0;
				snprintf(filename, sizeof(filename) - 1, PATH_TEMP "/tmpcert.ca%d", ++count);
				setvar(p_start, "_data_");
				save_var_as_file(filename, "_data_");
				*p_end = '\n';
			}

			p_start = p_end;
		}

		// Zertifikat in Zertifikatsspeicher
		if(NULL != getvar("keypassword"))
		    argv_insert[2] = strdup(getvar("keypassword"));

		if(exec_command(NULL, &text_out, &text_err, argv_insert, NULL)) {
			if(NULL != text_err)
				set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv_insert[0], text_err);
			return EXIT_FAILURE;
		}

		if(NULL != argv_insert[2])
		    free(argv_insert[2]);

		certcertlist();
		P_SITE = "certcertlist";
	}

	return EXIT_SUCCESS;
}

