/***************************************************************************
 *   Copyright (C) 2008 by msm net ingenieurbuero meissner   *
 *   service@msm-net.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <dirent.h>

#include "main.h"
#include "CERT.h"


extern int certreqlist(void);
int certreqnew_gen_local(void);
int certreqnew_gen_sc(void);
int certreqnew_gen_sc_get_keyids(char **output);


int certreqnew(void) {
	char *p_start;

    // set select/check Variables
	if(NULL != getvar("localsmart"))
		setvar("checked", "localsmart_%s", getvar("localsmart"));
	else
		setvar("checked", "localsmart_local");

	if(NULL != getvar("keysize"))
		setvar("selected", "keysize_%s", getvar("keysize"));

	if(NULL != getvar("smartkeysize"))
		setvar("selected", "smartkeysize_%s", getvar("smartkeysize"));


	if(NULL != getvar("execute")) {

	// check input for key-generation
		if(NULL == getvar("localsmart") || (strcmp(getvar("localsmart"), "local") && strcmp(getvar("localsmart"), "smart"))) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 75);
			return EXIT_FAILURE;
		}

        // check inputs for key if local
		if(!strcmp(getvar("localsmart"), "local")) {
			if(NULL == getvar("keysize")) {
				set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 76);
				return EXIT_FAILURE;
			}
			switch(atoi(getvar("keysize"))) {
				case 512:
				case 768:
				case 1024:
				case 1536:
				case 2048:
					break;				 
				default:
					set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 77);
					return EXIT_FAILURE;
			}
			if((NULL == getvar("password")) || (strlen(getvar("password")) < CERT_PASSWD_MIN) || (strlen(getvar("password")) > CERT_PASSWD_MAX)) {
				set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 78, CERT_PASSWD_MIN, CERT_PASSWD_MAX);
				return EXIT_FAILURE;
			}
			if((NULL == getvar("password")) || (NULL == getvar("password1")) || strcmp(getvar("password"), getvar("password1"))) {
				set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 79);
				return EXIT_FAILURE;
			}
		}
		// check inputs for key if on smartcard
		else {
			if(NULL == getvar("smartkeysize")) {
				set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 80);
				return EXIT_FAILURE;
			}
			setvar(getvar("smartkeysize"), "keysize");
			switch(atoi(getvar("keysize"))) {
				case 512:
				case 768:
				case 1024:
					break;
				default:
					set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 81);
					return EXIT_FAILURE;
			}

		if((NULL == getvar("adminpassword")) || (strlen(getvar("adminpassword")) < SC_PASSWD_MIN) || (strlen(getvar("adminpassword")) > SC_PASSWD_MAX)) {
				set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 82, SC_PASSWD_MIN, SC_PASSWD_MAX);
				return EXIT_FAILURE;
			}
			if((NULL == getvar("userpassword")) || (strlen(getvar("userpassword")) < SC_PASSWD_MIN) || (strlen(getvar("userpassword")) > SC_PASSWD_MAX)) {
				set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 83, SC_PASSWD_MIN, SC_PASSWD_MAX);
				return EXIT_FAILURE;
			}
		}

		if(NULL == getvar("country") || strlen(getvar("country")) != 2 || check_string_for_chars(getvar("country"), CS_CHARS, 1)) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 84);
			return EXIT_FAILURE;
		}

		if(NULL == getvar("province") || check_string_for_chars(getvar("province"), CS_CHARS "- ", 1)) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 85);
			return EXIT_FAILURE;
		}

		if(NULL == getvar("city") || check_string_for_chars(getvar("city"), CS_CHARS CS_NUMBERS "- ", 1)) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 86);
			return EXIT_FAILURE;
		}

		if(NULL == getvar("organisation") || check_string_for_chars(getvar("organisation"), CS_CHARS CS_NUMBERS "- ", 1)) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 87);
			return EXIT_FAILURE;
		}

		if(NULL == getvar("unit") || check_string_for_chars(getvar("unit"), CS_CHARS CS_NUMBERS "- ", 0)) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 88);
			return EXIT_FAILURE;
		}

		if(NULL == getvar("commonname") || check_string_for_chars(getvar("commonname"), CS_CHARS CS_NUMBERS "._- ", 1)) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 89);
			return EXIT_FAILURE;
		}

		if(NULL == getvar("email") || check_string_for_chars(getvar("email"), CS_CHARS CS_NUMBERS "@.-_", 1)) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 90);
			return EXIT_FAILURE;
		}
		else {
			if(NULL == (p_start = index(getvar("email"), '@')) || NULL != index(p_start + 1, '@')) {
				set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 90);
				return EXIT_FAILURE;
			}
		}

		if(!strcmp(getvar("localsmart"), "local")) 
			certreqnew_gen_local();
		else
			certreqnew_gen_sc();
		
		// Logeintrag
		syslog(LOG_DAEMON | LOG_INFO, "User: %s, Action: create new certificate request %s", "USRNAME", getvar("commonname"));

		// in Liste wchseln
		certreqlist();
		P_SITE = "certreqlist";
	}
	else
		syslog(LOG_DAEMON | LOG_INFO, "User: %s, Action: show page for new certificate request", "USRNAME");	
	
	return EXIT_SUCCESS;
}


/*
############################################################
*/
int certreqnew_gen_local(void) {
	char text_in[2048], *text_out, *text_err;
	char *argv_create_req[] = { BIN_OPENSSL, NULL };
	char *argv_insert[] = { BIN_INSERTCERTS, PATH_TEMP "/tmpcert", getvar("password"), NULL };

    // Request erstellen
	snprintf(text_in, sizeof(text_in) - 1, "req -config " CONF_OPENSSL " -new -newkey rsa:%i -keyout " PATH_TEMP "/tmpcert.key -out " PATH_TEMP "/tmpcert.req -passout pass:%s -subj \"/C=%s/ST=%s/L=%s/O=%s/OU=%s/CN=%s/emailAddress=%s\"", \
		atoi(getvar("keysize")), getvar("password"), \
		getvar("country"), getvar("province"), getvar("city"), getvar("organisation"), \
		getvar("unit"), getvar("commonname"), getvar("email"));
		
	if(exec_command(text_in, &text_out, &text_err, argv_create_req, NULL)) {
		if(NULL != text_err)
			set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv_create_req[0], text_err);
		return EXIT_FAILURE;
	}
	printf("%s<br><br>%s", text_out, text_err);

	// Request in Zertifikatsspeicher
	if(exec_command(NULL, &text_out, &text_err, argv_insert, NULL)) {
		if(NULL != text_err)
			set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv_insert[0], text_err);
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;

}


/*
############################################################
*/
int certreqnew_gen_sc(void) {
	char text_in[2048], *text_out, *text_err;
	char key[32], sopin[64], pin[64];
	char *argv_create_key[] = { BIN_PKSC15INIT, "--generate-key", key, "--auth-id", "01", "--so-pin", sopin, "--pin", pin, NULL };
	char *argv_create_req[] = { BIN_OPENSSL, NULL };
	char *argv_insert[] = { BIN_INSERTCERTS, PATH_TEMP "/tmpcert", getvar("password"), NULL };
	char *argv_readcard[] = { BIN_READSMARTCARD, CERT_PATH_SC, NULL };
	char *idlist_new, *idlist_old;
	int keycount = 0, keyid = 0;

	// check card
	if((keycount = certreqnew_gen_sc_get_keyids(&idlist_old)) < 0 || keycount >= 4) {
		set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 91);
		return EXIT_FAILURE;
	}
	
	snprintf(key, sizeof(key) - 1, "rsa/%d", atoi(getvar("keysize")));
	snprintf(pin, sizeof(pin) - 1, "%s", getvar("userpassword"));
	snprintf(sopin, sizeof(sopin) - 1, "%s", getvar("adminpassword"));

	if(exec_command(NULL, &text_out, &text_err, argv_create_key, NULL)) {
		if(NULL != text_err)
			set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv_create_key[0], text_err);
		return EXIT_FAILURE;
	}

	if(keycount == certreqnew_gen_sc_get_keyids(&idlist_new)) {
		set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 92);
		return EXIT_FAILURE;
	}

	if(NULL != idlist_new) {
		if(NULL != idlist_old)
			keyid = atoi(idlist_new + strlen(idlist_old));
		else
			keyid = atoi(idlist_new);
	}

	if(!keyid) {
		set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 93);
		return EXIT_FAILURE;
	}

	// Request erstellen	
	snprintf(text_in, sizeof(text_in) - 1,
		"engine dynamic -pre SO_PATH:" PATH_PKCS11ENGINE " -pre ID:pkcs11 -pre LIST_ADD:1 -pre LOAD\n" \
		"req -config " CONF_OPENSSL " -engine pkcs11 -new -key id_%d -keyform engine -out " PATH_TEMP "/tmpcert.req -subj \"/C=%s/ST=%s/L=%s/O=%s/OU=%s/CN=%s/emailAddress=%s\"\n" \
		"%s\n", \
		keyid, getvar("country"), getvar("province"), getvar("city"), \
		getvar("organisation"),getvar("unit"), getvar("commonname"), getvar("email"), \
		getvar("userpassword"));
//printf("<br>%s<br>\n", text_in);

	if(		exec_command(text_in, &text_out, &text_err, argv_create_req, NULL)
		|| 	(NULL != text_err && strcmp(text_err, "engine \"pkcs11\" set.\nSmartCard PIN: \n"))
	) {
//printf("<br>OUT (ERR): %s<br>\n", text_out);
//printf("<br>ERR (ERR): %s<br>\n", text_err);
		if(NULL != text_err)
			set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv_create_req[0], text_err);
		return EXIT_FAILURE;
	}

//printf("<br>OUT: %s<br>\n", text_out);
//printf("<br>ERR: %s<br>\n", text_err);

	// Request in Zertifikatsspeicher
	if(exec_command(NULL, &text_out, &text_err, argv_insert, NULL)) {
		if(NULL != text_err)
			set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv_insert[0], text_err);
		return EXIT_FAILURE;
	}

	// Smartcard erneut einlesen
	if(exec_command(NULL, &text_out, &text_err, argv_readcard, NULL)) {
		if(NULL != text_err)
			set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv_readcard[0], text_err);
		return EXIT_FAILURE;
	}


	return EXIT_SUCCESS;
}


/*
############################################################
*/
int certreqnew_gen_sc_get_keyids(char **output) {
	char *text_out, *text_err;
	char *argv_list_keys[] = { BIN_PCKS15TOOL, "--list-keys", NULL };
	char idlist[1024] = "", *p_start, *p_end;
	int keycount = 0;

	// check card
	if(exec_command(NULL, &text_out, &text_err, argv_list_keys, NULL)) {
		if(NULL != text_err)
			set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv_list_keys[0], text_err);
		return EXIT_FAILURE;
	}

	if(NULL != text_err)
		free(text_err);

	if(NULL != (p_start = text_out)) {
		while(NULL != (p_start = strstr(p_start, "[Private Key]"))) {
			++keycount;
			if(		NULL != (p_start = strstr(p_start, "\tID"))
			 	&&	NULL != (p_start = index(p_start, ':'))
			 	&&	NULL != (p_end = index(++p_start, '\n'))
			) {
				*p_end = 0;
				if(strlen(p_start) + strlen(idlist) < sizeof(idlist) - 1)
					strcat(idlist, p_start);
				p_start = ++p_end;
			}
		}
	}

	if(strlen(idlist))
		*output = strdup(idlist);
	else
		*output = NULL;

	return keycount;

}
