/***************************************************************************
 *   Copyright (C) 2008 by msm net ingenieurbuero meissner   *
 *   service@msm-net.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <signal.h>

#include "main.h"
#include "CERT.h"

int certsmartcard_pass(char **args, char *info);

int certsmartcard(void) {
	char *text_out, *text_err;
	char *argv_init1[] = { BIN_PKSC15INIT, "--erase-card", "--create-pkcs15", NULL };
	char *argv_init2[] = { BIN_PKSC15INIT, "--store-pin", "--auth-id", "01", "--label", "\"User\"", NULL };
	char *argv_readcard[] = { BIN_READSMARTCARD, CERT_PATH_SC, NULL };
	char *argv_detail[] = { BIN_PCKS15TOOL, "--list-public-keys", "--list-keys", "--list-certificates", "--list-pins", NULL };


	 if(NULL != getvar("execute")) {
		// Passwoerter pruefen
		if(NULL == getvar("adminpassword") || NULL == getvar("userpassword")) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 116);
			return EXIT_FAILURE;
		}

		if(NULL == getvar("newadminpassword") || NULL == getvar("newuserpassword")) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 117);
			return EXIT_FAILURE;
		}

		if(NULL == getvar("newadminpassword1") || strcmp(getvar("newadminpassword"), getvar("newadminpassword1"))) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 118);
			return EXIT_FAILURE;
		}

		if(NULL == getvar("newuserpassword1") || strcmp(getvar("newuserpassword"), getvar("newuserpassword1"))) {
			set_error_multilang(__FUNCTION__, __FILE__, __LINE__, 119);
			return EXIT_FAILURE;
		}
		addvar("\n", "adminpassword");
		addvar("\n", "userpassword");
		addvar("\n", "newadminpassword");
		addvar("\n", "newuserpassword");

printf("<br><br>ERASE:<br>\n");

		if(certsmartcard_pass(argv_init1, "Fehler beim Loeschen der Smartkarte"))
			return EXIT_FAILURE;

		setvar(getvar("newadminpassword"), "adminpassword");
		setvar(getvar("newuserpassword"), "userpassword");
printf("<br><br>AUTHID:<br>\n");		

		if(certsmartcard_pass(argv_init2, "Fehler beim Initialisieren der Smartkarte"))
			return EXIT_FAILURE;
		
		// Smartcard erneut einlesen
		if(exec_command(NULL, &text_out, &text_err, argv_readcard, NULL)) {
			if(NULL != text_err)
				set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv_readcard[0], text_err);
			return EXIT_FAILURE;
		}
	}

	// Smartcard-Details lesen
	if(exec_command(NULL, &text_out, &text_err, argv_detail, NULL)) {
		if(NULL != text_err)
			set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", argv_detail[0], text_err);
		return EXIT_FAILURE;
	}

	if(NULL != text_out)
		setvar(text_out, "text");

	syslog(LOG_DAEMON | LOG_INFO, "User: %s, Action: delete smartcard", "USRNAME");

	return EXIT_SUCCESS;
}

/*
############################################################
*/
int certsmartcard_pass(char **argv, char *info) {
	int stream_in[2], stream_out[2], stream_err[2];
	int pid, tmp_status, is_ret = 0, status = 255;
	int len_out = 0, lend_out = 0, len_err = 0, lend_err = 0;
	char buffer_out[1024] = {0}, buffer_err[1024] = {0};
	char *last = NULL;

	// Pipes
	if(0 != pipe(stream_in)) {
		set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n can't create pipe for stdin", info);
		return EXIT_FAILURE;
	}

	if(0 != pipe(stream_out)) {
		set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n can't create pipe for stdout", info);
		return EXIT_FAILURE;
	}

	if(0 != pipe(stream_err)) {
		set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n can't create pipe for stderr", info);
		return EXIT_FAILURE;
	}
	
	fcntl(stream_out[0], F_SETFL, O_NONBLOCK);
	fcntl(stream_err[0], F_SETFL, O_NONBLOCK);

	pid = fork();
	if(pid < 0) {
		set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n can't fork(%s)", info, *argv);
		exit(255);
	}

	if(!pid) {
		close(stream_in[1]);
		close(stream_out[0]);
		close(stream_err[0]);
	
		dup2(stream_in[0], 0);
		dup2(stream_out[1], 1);
		dup2(stream_err[1], 2);

		for(pid = 3; pid < 255; pid ++)
			close(pid);

		execv(argv[0], argv);
		exit(255);
	}

	close(stream_in[0]);
	close(stream_out[1]);
	close(stream_err[1]);
	
	while(!is_ret) {

		if(pid == waitpid(pid, &tmp_status, WNOHANG)) {
			is_ret = 1;
			status = tmp_status;
		}

//		lend_err = read(stream_err[0], buffer_err + len_err, sizeof(buffer_err) - len_err);
//		lend_out = read(stream_out[0], buffer_out + len_out, sizeof(buffer_out) - len_out);
		lend_err = read(stream_out[0], buffer_err + len_err, sizeof(buffer_err) - len_err);
		lend_out = read(stream_err[0], buffer_out + len_out, sizeof(buffer_out) - len_out);

		if(lend_err < 0 && lend_out < 0)
//			usleep(1000);
			usleep(500000);

		if(lend_err > 0)
			len_err += lend_err;

		if(lend_out > 0)
			len_out += lend_out;

		if(len_out > 2 && (*(buffer_out + len_out - 1) == ':' || *(buffer_out + len_out - 2) == ':')) {
 			*(buffer_out + len_out) = 0;

 			if(NULL != strstr(buffer_out, "verify")) {
//				 last = last;
			} else
				if(NULL != strstr(buffer_out, "PUK")) {
					last = "\n";
				} else
					if(NULL != strstr(buffer_out, "Security ")) {
						if(NULL != index(buffer_out, '['))
							last = getvar("adminpassword");
						else
				 			last = getvar("newadminpassword");
					} else {
						if(NULL != strstr(buffer_out, "New"))
							last = getvar("newuserpassword");
						else
				 			last = getvar("userpassword");
					}
//printf("<br>Frage:   '%s'\n", buffer_out);
//printf("<br>Antwort: '%s'\n", last);
			*buffer_out = 0;
			len_out = 0;
			write(stream_in[1], last, strlen(last));
			
		}
	}
	close(stream_in[1]);
	close(stream_out[0]);
	close(stream_err[0]);
//printf("<br>stdout: %s<br>\n", buffer_out);
//printf("<br>STAT: %d<br>\n", WEXITSTATUS(status));

	if(WEXITSTATUS(status))
		set_error(__FUNCTION__, __FILE__, __LINE__, "%s:\n%s", info, buffer_out);

	return WEXITSTATUS(status);

}
