/***************************************************************************
 *   Copyright (C) 2008 by msm net ingenieurbuero meissner   *
 *   service@msm-net.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <dirent.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <signal.h>

#include "config.h"
#include "errors.h"


int setvar(char *value, char *fmt, ...);
int setvar_direct(char *value, char *fmt, ...);
char *getvar(const char *fmt, ...);
int addvar(char *value, char *fmt, ...);
int getvars_from_file(char **pp_varname, char *p_filename);
void clear_all_vars(void);
char *str_upper(char *str);
char *str_lower(char *str);
int parse_multiform(void);
char *get_restlines_of_file(char *p_filename, char **linebeginn);
int read_file_in_var(char *p_filename, char *p_varname);
int save_var_as_file(char *p_filename, char *p_varname);
void set_error(const char *name, char *file, int line, const char *fmt, ...);
void set_error_multilang(const char *function, char *file, int line, const int number, ...);
void cleartmp(void);
void sort_optionlist(char *options);
int check_ip_fqdn(char *network, int must_ip, int addr_family);
int check_string_for_chars(char *string, char *allowed, int must);
int checktime_cert(char *certfile, unsigned int seconds);
int exec_command(char *textin, char **textout, char **texterr, char **argv, char **more_in);
/**/

static void do_nothing(void);
void logout(void);

int certreqlist(void);
int certreqnew(void);
int certreqexportview(void);
int certreqsign(void);
int certreqdel(void);
int certreqimport(void);

int certcertlist(void);
int certcertexportview(void);
int certcertexport(void);
int certcertimport(void);
int certcertdel(void);
int certcrl(void);

int certsmartcard(void);

int viewopensslconfig(void);

int msmcertadminmanual(void);


struct_siteindex siteindex[] = {
	{ "logout", &logout},
	{ "gpl", &do_nothing },
	{ "versionsinfo", &do_nothing },
	{ "mainmenu", &do_nothing },
	{ "certmenu", &do_nothing },
		{ "certreqlist", &certreqlist },
		{ "certreqnew", &certreqnew },
		{ "certreqexportview", &certreqexportview },
		{ "certreqsign", &certreqsign },
		{ "certreqdel", &certreqdel },
		{ "certreqimport", &certreqimport },
		{ "certcertlist", &certcertlist },
		{ "certcertexportview", &certcertexportview },
		{ "certcertexport", &certcertexport },
		{ "certcertimport", &certcertimport },
		{ "certcertdel", &certcertdel },
		{ "certsmartcard", &certsmartcard },
		{ "certcrl", &certcrl },
	{ "paketmenu", &do_nothing },
		{ "opensslmanual", &do_nothing },
		{ "viewopensslconfig", &viewopensslconfig },
		{ "pcscmanual", &do_nothing },
		{ "msmcertadminmanual", &msmcertadminmanual },
	{ NULL }
};

struct_vars VARS[_VARS_MAX_NUMBERS_ + 1] = {
	{ "", NULL }
};

char *P_SITE;
int SITE_INDEX = -1;
char *P_POST_VARS;
char ERROR_BUF[10240] = { 0 };
int SET_ERROR_NR = 0;
int RIGHTS_OK = 0;
int index_language = 0;

/*
############################################################
*/
int main(int argc, char *argv[])
{
	char site_function_not_found[] = "error_function_not_found";
	char site_site_not_found[] = "error_site_not_found";
	char site_mainmenu[] = "mainmenu";
	char to_decode[_VARS_NAME_MAX_LENGTH_ + 1] = {0};
	char filename[500];
	char *p_html = NULL;
	char *p_post_vars;
	char *p_varname, *p_value, *p_value1, *p_value2, *p_start, *p_end, *p_decode;
	char *p_buffer, *p_sitename, *p_rights;
	int counter = 0;
	void (*function)() = NULL;
	long size;
	FILE *p_file;
	struct stat file_info;
	char *conf_language = NULL;

	setvbuf(stdout, NULL, _IONBF, 0);
	stderr = stdout;
	setvbuf(stderr, NULL, _IONBF, 0);

	openlog("CertAdm", LOG_NDELAY | LOG_PID, LOG_DAEMON);

///////////////
//printf("Cache-Control: no-cache, must-revalidate\n"); // HTTP/1.1
//printf("Content-type: text/html\n");
//printf("\n");
//printf("%s<br><br>\n\n", getenv("CONTENT_TYPE"));
//////////////


	//get posted variables
	if((NULL != (p_post_vars = getenv("CONTENT_LENGTH"))) && (strlen(p_post_vars))) {
		// read from stdin into memory
		size = atoi(p_post_vars);
		p_post_vars = malloc(size + 2);
		P_POST_VARS = malloc(size + 1);
		fread(p_post_vars, 1, size, stdin);
		memcpy(P_POST_VARS, p_post_vars, size);
		*(p_post_vars + size) = '&';
		*(p_post_vars + size + 1) = 0;

//printf("%s<br><br>\n\n", p_post_vars);

		// get varname an value for every entry
		if(!strcmp(getenv("CONTENT_TYPE"), "application/x-www-form-urlencoded")) {
			p_varname = p_post_vars;
			while(NULL != (p_end = index(p_varname, '&'))) {
				*(p_end) = 0;
				if(NULL != (p_value = index(p_varname, '=')))
					*(p_value++) = 0;
				// decode application/x-www-form-urlencoded to plain
				p_decode = p_value;
				while(NULL != (p_decode = index(p_decode, '+'))) {
					*(p_decode++) = ' ';
				}
				p_decode = p_value;
				while(NULL != (p_decode = index(p_decode, '%'))) {
					memcpy(to_decode, p_decode + 1, 2);
					*(p_decode++) = (unsigned char)strtol(to_decode, NULL, 16);
					strcpy(p_decode, p_decode + 2);
				}
				// check for '\r'
				p_value2 = strdup(p_value);
				p_value1 = p_value2;
				while(NULL != (p_value1 = index(p_value1, '\r'))) {
					memmove(p_value1, p_value1 + 1, strlen(p_value1));
				}
				// set variables in global VARS
				setvar(p_value2, p_varname);
				free(p_value2);
				p_varname = p_end + 1;
			}
		}
		else {
			if(NULL != strstr(getenv("CONTENT_TYPE"), "multipart/form-data")) {
				parse_multiform();
			}
			else {
				//printf("ERROR: POSTED FORM is NOT multipart/form-data or application/x-www-form-urlencoded");
				syslog(LOG_DAEMON | LOG_INFO, "User: %s, posted form is NOT multipart/form-data or application/x-www-form-urlencoded - IGNORE", "USRNAME");
				goto _no_access;
			}
		}
		// free memory
		free(p_post_vars);
	}

	// set language
	index_language = 0;
	if(NULL == getvar("language") || NULL == (conf_language = strdup(getvar("language"))))
		conf_language = DEFAULTLANGUAGE;
	if(!strcmp(conf_language, "english"))
		index_language = 1;
	else 
		if(strcmp(conf_language, "german"))
			conf_language = DEFAULTLANGUAGE;

	//try to get function for site
	counter = 0;
	if(NULL != (P_SITE = getenv("QUERY_STRING")) && strlen(P_SITE)) {
		P_SITE = strdup(getenv("QUERY_STRING"));
		if(NULL != index(P_SITE, '&'))
			*index(P_SITE, '&') = 0;
		while(siteindex[counter].site != NULL) {
//printf("%s -- %s: %u<br>", siteindex[counter].site, P_SITE, siteindex[counter].function);
			if(!strcmp(siteindex[counter].site, P_SITE)) {
				function = siteindex[counter].function;
				SITE_INDEX = counter;
				break;
			}
			++counter;
		}
		if(NULL == function) {
			P_SITE = site_function_not_found;
			syslog(LOG_DAEMON | LOG_INFO, "User: %s, access to a unknown function - IGNORE", "USRNAME");
			goto _no_access;
		}
	}
	else
		P_SITE = site_mainmenu;

	// loesche tmp-verzeichnis
	if(!strcmp(P_SITE, "mainmenu"))
		cleartmp();

/*
	// check rights
	SITE_INDEX = counter;
	if((siteindex[counter].rights & LOCALADMIN_READ) < LOCALADMIN_READ) { 
		if(NULL != getvar("LASTPAGE"))
			P_SITE = ("LASTPAGE");
		else
			P_SITE = "mainmenu";
		goto _no_access;
	}
*/

	//function call
	if(NULL != function) {
		(*function)();
		sync();
	}

_no_access:
	setvar(ERROR_BUF, "ERROR");
	if(NULL != P_SITE)
		setvar(P_SITE, "LASTPAGE");

//open html-site
_site_not_found:
	sprintf(filename, "languages/%s/%s.html", conf_language, P_SITE);
	if(NULL != (p_file = fopen(filename, "r"))) {
		fseek(p_file, 0L, SEEK_END);
		size = ftell(p_file);
		rewind(p_file);
		if(NULL != (p_html = malloc(size + 1))) {
			if(!fread(p_html, 1, size, p_file))
				p_html = NULL;
			else
				*(p_html + size) = 0;
		}
		fclose(p_file);
	}
	else {
		if(P_SITE != site_site_not_found) {
			P_SITE = site_site_not_found;
			goto _site_not_found;
		}
	}

	// set language
	setvar(conf_language, "language");

	//replace variables in html-text and put it out
	//if(function != &logout) {
	printf("Cache-Control: no-cache, must-revalidate\n");
	printf("Content-type: text/html\n");
	printf("\n");
	//}

	if(NULL != p_html) {
		p_varname = p_html;
		p_start = p_html;
		while(NULL != (p_varname = index(p_varname, '~'))) {
			if(NULL != (p_end = index(p_varname + 1, '~'))) {
				if(((size = (p_end - p_varname)) < (_VARS_NAME_MAX_LENGTH_ + 1)) && size > 0){
					memcpy(to_decode, p_varname + 1, size -1);
					to_decode[size - 1] = 0;
					*(p_varname) = 0;
					printf("%s", p_start);
					if(NULL != (p_value = getvar(to_decode)))
						printf("%s", p_value);
					p_varname = p_end;
					p_start = p_varname +1;
				}
			}
			++p_varname;
		}
        	printf("%s", p_start);
		free(p_html);
	}
	else
		printf("%s", "FILE NOT FOUND");

	exit(EXIT_SUCCESS);
}


/*
############################################################
*/
void set_error(const char *name, char *file, int line, const char *fmt, ...) {
	char buf1[4200] = { 0 };
	char buf2[4096] = { 0 };
	va_list args;
	va_start (args, fmt);
	int i, offset;

	snprintf(buf1, sizeof(buf1), "%s (%s:%d)\n", name, file, line);
	vsnprintf(buf2, sizeof(buf2), fmt, args);
	strncat(buf1, buf2, sizeof(buf1) - 50);
	offset = 0;

	for(i = 0; i < strlen(buf1) && (i + offset) < sizeof(buf2) - 1; i++)
		switch(buf1[i]) {
			case '\n':
				buf2[i + offset] = '\\';
				buf2[i + offset +1] = 'n';
				offset += 1;
				break;
			case '\t':
				buf2[i + offset] = '\\';
				buf2[i + offset +1] = 't';
				offset += 1;
				break;				
			case '\"':
				buf2[i + offset] = '\\';
				buf2[i + offset +1] = '\"';
				offset += 1;
				break;
			case '\\':
				buf2[i + offset] = '\\';
				buf2[i + offset +1] = '\\';
				offset += 1;
				break;
			default:
				buf2[i + offset] = buf1[i];
				break;
		}
	buf2[i + offset] = 0;
	
	snprintf(buf1, sizeof(buf1), "msg%d=\"%s\";setTimeout(\"alert(msg%d)\", 1000);\n", SET_ERROR_NR, buf2, SET_ERROR_NR);
	SET_ERROR_NR++;
	strcat(ERROR_BUF, buf1);

}

/*
############################################################
*/
void set_error_multilang(const char *name, char *file, int line, int number, ...) {
	t_ERRTEXT *Errtexts = ERRTEXT;
	char buf1[4200] = { 0 };
	char buf2[4096] = { 0 };
	int i, offset;
	const char *fmt = "unkown error - no message text found";
	va_list args;
	va_start (args, number);

	snprintf(buf1, sizeof(buf1), "%s (%s:%d)\n", name, file, line);
	while(Errtexts->MsgNr)
		if(Errtexts->MsgNr == number) {
			if(!strcmp(getvar("language"), "german"))
				fmt = Errtexts->german;
			else if(!strcmp(getvar("language"), "english"))
				fmt = Errtexts->english;
			break;
		}
		else
			Errtexts++;
	if(Errtexts->MsgNr)
		vsnprintf(buf2, sizeof(buf2), fmt, args);	
	else
		snprintf(buf2, sizeof(buf2), fmt);
	strncat(buf1, buf2, sizeof(buf1) - 50);
	offset = 0;

	for(i = 0; i < strlen(buf1) && (i + offset) < sizeof(buf2) - 1; i++)
		switch(buf1[i]) {
			case '\n':
				buf2[i + offset] = '\\';
				buf2[i + offset +1] = 'n';
				offset += 1;
				break;
			case '\t':
				buf2[i + offset] = '\\';
				buf2[i + offset +1] = 't';
				offset += 1;
				break;				
			case '\"':
				buf2[i + offset] = '\\';
				buf2[i + offset +1] = '\"';
				offset += 1;
				break;
			case '\\':
				buf2[i + offset] = '\\';
				buf2[i + offset +1] = '\\';
				offset += 1;
				break;
			default:
				buf2[i + offset] = buf1[i];
				break;
		}
		buf2[i + offset] = 0;
	
		snprintf(buf1, sizeof(buf1), "msg%d=\"%s\";setTimeout(\"alert(msg%d)\", 1000);\n", SET_ERROR_NR, buf2, SET_ERROR_NR);
		SET_ERROR_NR++;
		strcat(ERROR_BUF, buf1);
}


/*
############################################################
*/

void logout(void) {

	printf("WWW-Authenticate: Basic realm=\"cgi-bin\"\n");
	printf("%s 401 Unauthorized\n", getenv("SERVER_PROTOCOL"));
/*	
	printf("Server: thttpd/2.23beta1 26may2002\n");
	printf("Date: Wed, 31 Jul 2003 17:35:20 GMT\n");
	printf("Last-Modified: Wed, 31 Jul 2003 17:35:20 GMT\n");
	printf("Accept-Ranges: bytes\n");
	printf("Connection: close\n");
*/
	return;
}


/*
############################################################
*/
int setvar(char *value, char *fmt, ...) {
	int size, counter = 0;
	char varname[_VARS_NAME_MAX_LENGTH_ + 1];
	va_list args;
	va_start(args, fmt);

	if(NULL == value)
		value = "";
	vsnprintf(varname, sizeof(varname), fmt, args);

	if(!(size = strlen(varname)) || (size > _VARS_NAME_MAX_LENGTH_)) {
// VARIABLEN NAME ZU LANG oder keine NAME
		return -2;
	}

	while(0 != *(VARS[counter].name)) {
		if(!strcmp(varname, VARS[counter].name)) {
			if(NULL != VARS[counter].value)
				free(VARS[counter].value);
			break;
		}
		if(++counter > _VARS_MAX_NUMBERS_) {
// ZU VIELE VARIABLEN
			return -1;
		}
	}
	strcpy(VARS[counter].name, varname);
	VARS[counter].value = malloc(strlen(value) + 1);

	if(0 == strlen(value))
		VARS[counter].value = NULL;
	else
		strcpy(VARS[counter].value, value);

	return 0;
}


/*
############################################################
*/
int setvar_direct(char *value, char *fmt, ...) {
	int size, counter = 0;
	char varname[_VARS_NAME_MAX_LENGTH_ + 1];
	va_list args;
	va_start(args, fmt);

	vsnprintf(varname, sizeof(varname), fmt, args);

	if(!(size = strlen(varname)) || (size > _VARS_NAME_MAX_LENGTH_)) {
// VARIABLEN NAME ZU LANG oder keine NAME
        	return -2;
	}

	while(0 != *(VARS[counter].name)) {
		if(!strcmp(varname, VARS[counter].name)) {
			if(NULL != VARS[counter].value)
				free(VARS[counter].value);
			break;
		}
		if(++counter > _VARS_MAX_NUMBERS_) {
// ZU VIELE VARIABLEN
			return -1;
		}
	}
	strcpy(VARS[counter].name, varname);
	VARS[counter].value = value;

	return 0;
}


/*
############################################################
*/
char *getvar(const char *fmt, ...) {	
	int counter = 0;
	char varname[_VARS_NAME_MAX_LENGTH_ + 1];
	va_list args;
	va_start(args, fmt);

	vsnprintf(varname, sizeof(varname), fmt, args);

	while((counter < _VARS_MAX_NUMBERS_) || (0 != *(VARS[counter].name))) {
		if(!strcmp(varname, VARS[counter].name)) {
			return VARS[counter].value;
		}
		++counter;
	}
	return NULL;
}


/*
############################################################
*/
int addvar(char *value, char *fmt, ...) {
	int size, counter = 0;
	char *p_value_old;
	char varname[_VARS_NAME_MAX_LENGTH_ + 1];
	va_list args;
	va_start(args, fmt);

	vsnprintf(varname, sizeof(varname), fmt, args);

	if(NULL == value || 0 == strlen(value))
		return 0;
	
 	if(!(size = strlen(varname)) || (size > _VARS_NAME_MAX_LENGTH_)) {
// VARIABLEN NAME ZU LANG oder keine NAME
		return -2;
	}

	while(0 != *(VARS[counter].name)) {
		if(!strcmp(varname, VARS[counter].name))
			break;
			
		if(++counter > _VARS_MAX_NUMBERS_) {
// ZU VIELE VARIABLEN
			return -1;
		}
	}
	strcpy(VARS[counter].name, varname);
	p_value_old = VARS[counter].value;

	if(p_value_old != NULL) {
		VARS[counter].value = malloc(strlen(p_value_old) + strlen(value) + 1);
		*(VARS[counter].value) = 0;
		strcat(VARS[counter].value, p_value_old);
		strcat(VARS[counter].value, value);
		free(p_value_old);
	}
	else {
		VARS[counter].value = strdup(value);
	}
	
	return 0;
}



/*
############################################################
*/
int getvars_from_file(char **pp_varname, char *filename) {
	char *p_buffer;
	char *p_varname;
	char *p_value_start, *p_value_end;
	int size;
	FILE *p_file;

// daten aus datei lesen
	if(NULL == (p_file = fopen(filename, "r"))) {
		return EXIT_FAILURE;
	}

//groesse ermitteln
	fseek(p_file, 0L, SEEK_END);
	size = (int) ftell(p_file);
	rewind(p_file);

//speicher reserviren;
	if(NULL == (p_buffer = malloc(size + 1))) {
		fclose(p_file);
		return EXIT_FAILURE;
	}

//datei in speicher lesen und schliessen
	if(!fread(p_buffer, size, 1, p_file)) {
		free(p_buffer);
		fclose(p_file);
		return EXIT_FAILURE;
	}
	fclose(p_file);
	*(p_buffer + size) = 0;

// daten holen
	while(NULL != (p_varname = *pp_varname)) {
		p_value_start = p_buffer;
		while(NULL != (p_value_start = strstr(p_value_start, p_varname))) {

			if(p_value_start != p_buffer && !((*(p_value_start - 1) == '\n') || (*(p_value_start - 1) == '\t'))) {
				p_value_start++;
				continue;
			}

			p_value_start += strlen(p_varname) +1;

			if(*(p_value_start -1) != '=' && *(p_value_start -1) != ' ')
				continue;

			if(*p_value_start == '\n')
				break;

			if(*p_value_start == '\n')
				break;

			if(NULL == (p_value_end = index(p_value_start, '\n'))) {
				setvar(p_value_start, p_varname);
			}

			else {
				*p_value_end = 0;
				setvar(p_value_start, p_varname);
				*p_value_end = '\n';
			}

			p_value_start = getvar(*pp_varname);

			while(NULL != (p_value_start = index(p_value_start, _NEWLINE_REPLACE_)))
				*p_value_start = '\n';
			break;
		}
		pp_varname++;
	}

	free(p_buffer);

	return EXIT_SUCCESS;
}


/*
############################################################
*/
void clear_all_vars(void) {
	int i;
	
	for(i =	0; i < _VARS_MAX_NUMBERS_; i++) {
		VARS[i].name[0] = 0;
		if(NULL != VARS[i].value)
			free(VARS[i].value);
		VARS[i].value = NULL;
	}
}


/*
############################################################
*/
static void do_nothing(void) {
	return;
}



/*
############################################################
*/
char *str_upper(char *str) {
	
	char *pointer;

	pointer = str = strdup(str);

	while(0 != *pointer) {
		*pointer = toupper(*pointer);
		pointer++;
	}

	return str;
}


/*
############################################################
*/
char *str_lower(char *str) {

	char *pointer;

	pointer = str = strdup(str);

	while(0 != *pointer) {
		*pointer = tolower(*pointer);
		pointer++;
	}

	return str;
}

/*
############################################################
*/


int parse_multiform(void) {
	char *p_stop;
	char end[] = "--\r\n";
	char *p_start = P_POST_VARS, *p_end;
	char *seperator;
	int seperator_size;
	int saveasfile;
	char filename[1024], buffer_1[2048], buffer_2[2048];
	char *p_varinhalt, *p_varname_start, *p_varname_end, varname[_VARS_NAME_MAX_LENGTH_];

	// setze Zeiger Ende
	p_stop = P_POST_VARS + atoi(getenv("CONTENT_LENGTH"));

	//suche trennung
	if(NULL == (p_end = memmem(p_start, p_stop - p_start + 1, "\r\n", 2))) {
		set_error(__FUNCTION__, __FILE__, __LINE__, "form ist not multipart/form-data\n");
		return EXIT_FAILURE;
	}

	seperator_size = p_end - p_start + 2;
	seperator = malloc(seperator_size);
	*seperator = '\r';
	*(seperator + 1) = '\n';
	memcpy(seperator + 2, p_start, seperator_size - 2);

	// Zeiger auf 1. Zeile
	p_start += seperator_size;

	// lese daten
	while(p_start != (p_end = memmem(p_start, p_stop - p_start + 1, end, sizeof(end))) && p_end != NULL) {
		p_start += 2;
		p_end = index(p_start, '\r');
		*p_end = 0;

		//hole varname
		if(NULL == (p_varname_start = strstr(p_start, "name=\"")))
			continue;

		p_varname_start += 6;
		if(NULL == (p_varname_end = strstr(p_varname_start, "\"")))
			continue;

		snprintf(varname, p_varname_end - p_varname_start + 1, "%s", p_varname_start);

		//pruefe ob als file speichern
		if(NULL != (p_varname_start = strstr(p_start, "filename=\""))) {
			saveasfile = 1;
			p_varname_start += 10;
			if(NULL == (p_varname_end = strstr(p_varname_start, "\"")))
				continue;
			snprintf(filename, p_varname_end - p_varname_start + 1, "%s", p_varname_start);
		}

		else
			saveasfile = 0;

		//suche anfang inhalt variable
		p_start = memmem(p_end , p_stop - p_end + 1, "\n\r\n", 3) + 3;

		//suche trennung
		if(NULL == (p_end = memmem(p_start, p_stop - p_start + 1, seperator, seperator_size))) {
			set_error(__FUNCTION__, __FILE__, __LINE__, "Seperator not found!\n");
			break;
		}


		// lesen
		if(saveasfile) {
			setvar(filename, varname);
			snprintf(buffer_1, sizeof(buffer_1), "%s_size", varname);
			snprintf(buffer_2, sizeof(buffer_2), "%d", p_end - p_start);
			setvar(buffer_2, buffer_1);
			p_varinhalt = malloc(p_end - p_start + 2);
			snprintf(buffer_1, sizeof(buffer_1), "%s_pointer", varname);
			setvar_direct(p_varinhalt, buffer_1);
			memcpy(p_varinhalt, p_start, p_end - p_start);
		}

		else {
			p_varinhalt = malloc(p_end - p_start + 2);
			snprintf(p_varinhalt, p_end - p_start + 1, "%s", p_start);
			setvar(p_varinhalt, varname);
			free(p_varinhalt);
		}

		p_start = p_end + seperator_size;
	}

	return EXIT_SUCCESS;

}


/*
############################################################
*/
char *get_restlines_of_file(char *p_filename, char **linebeginn) {
	FILE *p_file;
	struct stat file_info;
	char *p_buffer, *p_start, *p_end;

	if(NULL != (p_file = fopen(p_filename, "r"))) {

		if(fstat(fileno(p_file), &file_info))
			return NULL;
		p_buffer = malloc(file_info.st_size + 10);
		fread(p_buffer, file_info.st_size, 1, p_file);
		*(p_buffer + file_info.st_size) = 0;
		fclose(p_file);

		while(*linebeginn != NULL) {
			p_start = p_buffer;
			while(NULL != (p_start = strstr(p_start, *linebeginn)))  {
				if(p_start == p_buffer || *(p_start - 1) == '\n') {
					if(NULL != (p_end = index(p_start, '\n')))
						memmove(p_start, p_end + 1, strlen(p_end) + 1);
					else
						*p_start = 0;
				}
				else
					p_start++;
			}
			linebeginn++;
		}

		if(strlen(p_buffer))
			if(*(p_buffer + strlen(p_buffer) - 1) != '\n') {
				*(p_buffer + strlen(p_buffer) + 1) = 0;
				*(p_buffer + strlen(p_buffer)) = '\n';
			}

		p_start = p_buffer;

		while(NULL != (p_start = strstr(p_start, "\n\n")))
			memmove(p_start, p_start + 1, strlen(p_start) + 1);

		if(strlen(p_buffer) == 1)
			*p_buffer = 0;

		return p_buffer;

	}
	else
		return NULL;
}


/*
############################################################
*/
int read_file_in_var(char *p_filename, char *p_varname) {
	FILE *p_file;
	struct stat file_info;
	char *p_buffer;

	if(NULL != (p_file = fopen(p_filename, "r"))) {

		if(fstat(fileno(p_file), &file_info)) {
			set_error(__FUNCTION__, __FILE__, __LINE__, "read_file_in_var: error stat %s (%s)\n", p_filename, strerror(errno));
			return EXIT_FAILURE;
		}

		p_buffer = malloc(file_info.st_size + 1);
		fread(p_buffer, file_info.st_size, 1, p_file);
		*(p_buffer + file_info.st_size) = 0;
		fclose(p_file);

		setvar(p_buffer, p_varname);
		free(p_buffer);

		return EXIT_SUCCESS;
	}

	return EXIT_FAILURE;
}


/*
############################################################
*/
int save_var_as_file(char *p_filename, char *p_varname) {
	FILE *p_file;
	char *p_content;
	
	if(NULL == (p_file = fopen(p_filename, "w"))) {
			//set_error(__FUNCTION__, __FILE__, __LINE__, "save_var_as_file: error saveing file %s\n", p_filename);
			return EXIT_FAILURE;
	}

	if(NULL != (p_content = getvar(p_varname))) {
		while(NULL != (p_content = index(p_content, '\r')))
			memmove(p_content, p_content + 1, strlen(p_content));
		fprintf(p_file, "%s", getvar(p_varname));
	}

	fclose(p_file);

	return EXIT_SUCCESS;
}


/*
############################################################
*/
void cleartmp(void) {
	char dirname[] = PATH_TEMP;
	char buffer[1024];
	DIR *directory;
	struct dirent *directory_structur;

	if(NULL != (directory = opendir(dirname)))
		while(NULL != (directory_structur = readdir(directory)))

	if(*directory_structur->d_name != '.') {
		snprintf(buffer, sizeof(buffer), "%s/%s", dirname, directory_structur->d_name);
		if(0 > unlink(buffer))
			set_error(__FUNCTION__, __FILE__, __LINE__, "Failed to unlink %s (%s)", buffer, strerror(errno));
	}

	return;
}


/*
############################################################
*/
static unsigned char allgemein_itoa64[] =         /* 0 ... 63 => ascii - 64 */
        "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

static void allgemein_to64(register char *s, register long v, register int n) {
	while (--n >= 0) {
		*s++ = allgemein_itoa64[v&0x3f];
		v >>= 6;
	}
}


/*
############################################################
*/
int check_ip_fqdn(char *network, int must_ip, int addr_family) {
	union {
		struct in_addr ipv4;
		struct in6_addr ipv6;
	} ip;

	if(NULL == network)
	return EXIT_FAILURE;

	if((*network >= '0' && *network <= '9') || must_ip) {
		if(!inet_pton(addr_family, network, &ip)) 
			return EXIT_FAILURE;
	}

	else {
		while(*network) {
			if(		(*network >= 'A' && *network <= 'Z')
				|| 	(*network >= 'a' && *network <= 'z')
				||	(*network >= '0' && *network <= '9')
				||	*network == '.'
				||	*network == '-'
				||	*network == '_'
			)
				network++;
			else
				return EXIT_FAILURE;
		}
	}

	return EXIT_SUCCESS;
}

/*
############################################################
*/
int check_string_for_chars(char *string, char *allowed, int must) {
	char *p_allowed;

	if(NULL == string) {
		if(must)
			return EXIT_FAILURE;
		else
			return EXIT_SUCCESS;
	}
		
	while(*string) {
		p_allowed = allowed;

		while(*p_allowed)
			if(*string == *p_allowed)
				break;
			else
				p_allowed++;

		if(! *p_allowed) {
//printf("\n\n\n%x:%c<br>\n", *string, *string);
			return EXIT_FAILURE;
		}

		string++;
	}

	return EXIT_SUCCESS;
}

/*
############################################################
*/
void sort_optionlist(char *options) {
	char *new_options;
	char *ps_act, *pv_act, *pe_act, *ps_cmp, *pv_cmp, *pe_cmp;
	int cmp;

	if(NULL == options || !strlen(options))
		return;
		
	new_options = calloc(1, strlen(options) + 1);

	while(options == (ps_act = (ps_cmp = strstr(options, "<option")))) {

		if(NULL == (pe_cmp = index(ps_cmp + 1, '<'))) {
			strcat(new_options, ps_cmp);
			break;
		}

		while(NULL != (ps_act = strstr(ps_act + 1, "<option"))) {

			if(NULL == (pe_act = index(ps_act + 1, '<')))
				pe_act = ps_act + strlen(ps_act);

			if(NULL != (pv_act = index(ps_act, '>')) && NULL != (pv_cmp = index(ps_cmp, '>'))) {
				pv_act++;
				pv_cmp++;

				if(		(cmp = strncmp(pv_cmp, pv_act, (pe_act - pv_act < pe_cmp - pv_cmp) ? pe_act - pv_act : pe_cmp - pv_cmp)) > 0
					||	(!cmp && pe_act - ps_act < pe_cmp - ps_cmp)) {
					ps_cmp = ps_act;
					pe_cmp = pe_act;
				}
			}
		}

		strncat(new_options, ps_cmp, pe_cmp - ps_cmp);
		memmove(ps_cmp, pe_cmp, strlen(pe_cmp) +1);
	}

	memmove(options, new_options, strlen(new_options) + 1);
	free(new_options);
	
	return;
}


/*
############################################################
*/
int exec_command(char *text_in, char **text_out, char **text_err, char **argv, char **more_in) {
	int pid, tmp_status, is_ret = 0, status = 255;
	int stream_in[2], stream_out[2], stream_err[2], count_more_in = 0, point_more_in, *stream_more_in = NULL;
	int len_out = 0, lens_out = 0, len_err = 0, lens_err = 0;
	char buffer_out[1024], buffer_err[1024];
	off_t length;

	if(NULL != text_out)
		*text_out = NULL;

	if(NULL != text_err)
		*text_err = NULL;

	if(0 != pipe(stream_in)) {
		set_error(__FUNCTION__, __FILE__, __LINE__, "%s: can't create pipe for stdin", __FUNCTION__);
		return 255;
	}

	if(0 != pipe(stream_out)) {
		set_error(__FUNCTION__, __FILE__, __LINE__, "%s: can't create pipe for stdout", __FUNCTION__);
		return 255;
	}

	if(0 != pipe(stream_err)) {
		set_error(__FUNCTION__, __FILE__, __LINE__, "%s: can't create pipe for stderr", __FUNCTION__);
		return 255;
	}

	if(NULL != more_in) {

		while(NULL != *(more_in + count_more_in))
			count_more_in++;

		if(count_more_in) {

			if(NULL == (stream_more_in = malloc(sizeof(int *) * count_more_in * 2))) {
				set_error(__FUNCTION__, __FILE__, __LINE__, "%s: can't alloc mem for pipe(filedes) for more_in", __FUNCTION__);
				return 255;
			}

			for(point_more_in = 0; point_more_in < count_more_in; point_more_in++)
				if(0 != pipe(stream_more_in + point_more_in * 2)) {
					set_error(__FUNCTION__, __FILE__, __LINE__, "%s: can't create pipe for more_in[%d]", __FUNCTION__, point_more_in);
					return 255;
				}

		}
	}

	fcntl(stream_out[0], F_SETFL, O_NONBLOCK);
	fcntl(stream_err[0], F_SETFL, O_NONBLOCK);

	pid = fork();

	if(pid < 0) {
		set_error(__FUNCTION__, __FILE__, __LINE__, "%s: can't fork(%s)", __FUNCTION__, *argv);
		exit(255);
	}

	if(!pid) {
		dup2(stream_in[0], 0);
		dup2(stream_out[1], 1);
		dup2(stream_err[1], 2);

		for(point_more_in = 0; point_more_in < count_more_in; point_more_in++)
			dup2(*(stream_more_in + 2 * point_more_in), 3 + point_more_in);
		
		for(length = 3 + count_more_in; length < 255; length ++)
			close(length);

		execv(argv[0], argv);
		exit(255);
	}

	if(NULL != text_in)
		write(stream_in[1], text_in, strlen(text_in));

	close(stream_in[1]);

	for(point_more_in = 0; point_more_in < count_more_in; point_more_in++) {
		write(*(stream_more_in + 2 * point_more_in + 1), *(more_in + point_more_in), strlen(*(more_in + point_more_in)));
		close(*(stream_more_in + 2 * point_more_in + 1));
	}

	while(!is_ret || -1 != len_out || -1 != len_err) {
		if(pid == waitpid(pid, &tmp_status, WNOHANG)) {
			is_ret = 1;
			status = tmp_status;
		}
//printf("is_ret: %d<br>\n", is_ret);		
		if((len_out = read(stream_out[0], buffer_out, sizeof(buffer_out))) > 0 && NULL != text_out) {
			*text_out = realloc(*text_out, lens_out + len_out + 1);
			memcpy(*text_out + lens_out, buffer_out, len_out);
			lens_out += len_out;
		}
//printf("lens_out: %d, len_out: %d<br>\n", lens_out, len_out);		
		if((len_err = read(stream_err[0], buffer_err, sizeof(buffer_err))) > 0 && NULL != text_err) {
			*text_err = realloc(*text_err, lens_err + len_err + 1);
			memcpy(*text_err + lens_err, buffer_err, len_err);
			lens_err += len_err;
		}
//printf("lens_err: %d, len_err: %d<br>\n", lens_err, len_err);
		if(-1 == len_out && -1 == len_err)
			usleep(1000);
	}

	if(NULL != text_out && NULL != *text_out)
		*(*text_out + lens_out) = 0;

	if(NULL != text_err && NULL != *text_err)
		*(*text_err + lens_err) = 0;

	if(NULL != more_in)
		free(stream_more_in);

	close(stream_out[0]);
	close(stream_err[0]);

	close(stream_in[0]);
	close(stream_out[1]);
	close(stream_err[1]);
	
	return WEXITSTATUS(status);
}


/*
############################################################
*/
int fillin_vars(char *text_in, char *varname_out, char mark_begin, char mark_end) {
	char *p_var_start, *p_var_end, *p_text_start, *p_value, *p_newline;
	char varname[_VARS_NAME_MAX_LENGTH_ + 1] = {0};
	long size;

	if(NULL == text_in || NULL == varname_out)
		return EXIT_FAILURE;

	p_var_start = text_in;
	p_text_start = text_in;

	while(NULL != (p_var_start = index(p_text_start, mark_begin))) {

		if(NULL != (p_var_end = index(p_var_start + 1, mark_end))) {

			if(NULL != (p_newline = index(p_var_start, '\n')) && p_var_end > p_newline) {
				*p_newline = 0;
				addvar(p_text_start, varname_out);
				addvar("\n", varname_out);
				*p_newline = '\n';
				p_text_start = p_newline;
			}

			else {
				if(((size = (p_var_end - p_var_start)) < (_VARS_NAME_MAX_LENGTH_ + 1)) && size > 0){
					memcpy(varname, p_var_start + 1, size -1);
					*p_var_start = 0;
//printf("<br>%s<br>\n", p_text_start);				
					addvar(p_text_start, varname_out);
					*p_var_start = mark_begin;
					if(NULL != (p_value = getvar(varname)))
						addvar(p_value, varname_out);
				}
				else {
					*p_var_end = 0;
					addvar(p_text_start, varname_out);
					*p_var_end = mark_end;
				}
				p_text_start = p_var_end;
			}
		}

		else
			break;
		++p_text_start;
	}

	addvar(p_text_start, varname_out);

	return EXIT_SUCCESS;
}
